import json
from typing import Dict

from rest_framework import serializers

from providers.models import AnalyticsToken, Provider, ProviderSchema


class ProviderSchemaSerializer(serializers.ModelSerializer):
    definition = serializers.SerializerMethodField("_definition")

    def _definition(self, obj: ProviderSchema):
        return {
            "id": obj.provider.definition_id,
            "name": obj.provider.name,
            "nameI18n": obj.provider.nameI18n,
            "description": obj.provider.description,
            "descriptionI18n": obj.provider.descriptionI18n,
            "groups": obj.groups,
            "essential_verbs": obj.essential_verbs,
        }

    class Meta:
        model = ProviderSchema
        fields = "__all__"


class ProvidersSerializer(serializers.ModelSerializer):
    versions = serializers.SerializerMethodField("_versions")

    def _versions(self, obj):
        schemas = ProviderSchema.objects.filter(provider=obj).order_by("-created")
        serializer = ProviderSchemaSerializer(schemas, many=True)
        return serializer.data

    class Meta:
        model = Provider
        fields = ["id", "name", "versions","nameI18n","descriptionI18n"]


class UserVerbSerializer(serializers.Serializer):
    id = serializers.CharField()
    consented = serializers.BooleanField()
    objects = serializers.CharField()


class SaveUserConsentSerializer(serializers.Serializer):
    def get_verb_objects_lookup(self, verb: Dict) -> Dict:
        try:
            object_consents = {}
            objects_arr = json.loads(verb["objects"])
            for obj in objects_arr:
                object_consents[obj["id"]] = obj["consented"]
            return object_consents
        except json.JSONDecodeError:
            raise serializers.ValidationError(
                f"Invalid json for objects in verb {verb['id']}"
            )

    def validate(self, data):
        """
        Validate verbs with same id have consistent consent (including objects)
        """
        look_up = {}
        for verb in data["verbs"]:
            if look_up.get(verb["id"]) is None:
                look_up[verb["id"]] = {
                    "consented": verb["consented"],
                    "object_consents": self.get_verb_objects_lookup(verb),
                }
            else:
                if look_up[verb["id"]]["consented"] != verb["consented"]:
                    raise serializers.ValidationError(
                        f"Verb {verb['id']} contains ambiguous consent."
                    )
                if look_up[verb["id"]][
                    "object_consents"
                ] != self.get_verb_objects_lookup(verb):
                    raise serializers.ValidationError(
                        f"Verb {verb['id']} contains ambiguous object consents."
                    )

        return data

    providerId = serializers.PrimaryKeyRelatedField(
        queryset=Provider.objects.all(), required=False
    )
    providerSchemaId = serializers.PrimaryKeyRelatedField(
        queryset=ProviderSchema.objects.all(), required=False
    )
    verbs = serializers.ListSerializer(child=UserVerbSerializer())
