import { Component, ElementRef, OnDestroy, OnInit, ViewChild } from '@angular/core'
import { ToastrService } from 'ngx-toastr'
import {
    ApiService,
    AnalyticsToken,
    AvailableAnalyticTokensDef,
    AnalyticsTokenVerb} from '../../services/api.service'
import { Clipboard } from '@angular/cdk/clipboard'
import { animate, state, style, transition, trigger } from '@angular/animations'
import { MatDialog } from '@angular/material/dialog'
import { CreateTokenDialog } from 'src/app/dialogs/create-token-dialog/create-token-dialog'
import * as moment from 'moment'
import { DeleteDialog } from 'src/app/dialogs/delete-dialog/delete-dialog'
import { Subscription } from 'rxjs'
import { environment } from 'src/environments/environment'

type AnalyticsTokenId = number

@Component({
    selector: 'app-analytics-tokens',
    templateUrl: './analytics-tokens.component.html',
    styleUrls: ['./analytics-tokens.component.scss'],
    animations: [
        trigger('detailExpand', [
            state('collapsed', style({ height: '0px', minHeight: '0' })),
            state('expanded', style({ height: '*' })),
            transition('expanded <=> collapsed', animate('225ms cubic-bezier(0.4, 0.0, 0.2, 1)'))
        ])
    ]
})
export class AnalyticsTokensComponent implements OnInit, OnDestroy {
    analyticTokens: Array<AnalyticsToken> = []
    availableAnalyticTokenVerbs: Array<AvailableAnalyticTokensDef> = []
    displayedColumns: string[] = ['name', 'key', 'created', 'expires']
    columnsToDisplayWithExpand = [...this.displayedColumns, 'expand']
    expandedElement: AnalyticsToken | null = null
    createTokenSub?: Subscription
    deleteTokenSub?: Subscription
    Object = Object
    @ViewChild('fileInput') fileInputRef!:ElementRef;
    selectedFile : File | null  = null;
    
    constructor(
        private _apiService: ApiService,
        private _clipboard: Clipboard,
        private _toastrService: ToastrService,
        private _dialog: MatDialog
    ) {}

    ngOnInit(): void {
        this.getAnalyticsTokens()
        this.getAvailableVerbs()
    }

    ngOnDestroy(): void {
        this.createTokenSub?.unsubscribe()
        this.deleteTokenSub?.unsubscribe()
    }

    getAnalyticsTokens(): void {
        this._apiService.getAnalyticsTokens().subscribe((analyticTokens) => {
           this.analyticTokens = analyticTokens;
           this.analyticTokens.map((e) => {
             this.fillCanAccessOptions(e, this.analyticTokens)
        })
        })
    }

    getAvailableVerbs() : void 
    {
        this._apiService.getAnalyticTokensAvailableVerbs().subscribe((availableVerbDefinition) => {
            
            this.availableAnalyticTokenVerbs = availableVerbDefinition.map( def => {
                let seen : {[key : string] : boolean} = {}
                return ({...def, analyticTokenVerbs : def.analyticTokenVerbs.filter((item) => 
                    seen.hasOwnProperty(item.id) ? false : (seen[item.id] = true)
                )})
            })
        })
    }

    onCreateAnalyticsTokenClick(): void {
        const dialogRef = this._dialog.open(CreateTokenDialog, {
            width: '800px',
            data: {
                headerText: $localize`:@@createAnalyticsTokenDialogHeader:Create New Analytics Token`
            }
        })

        this.deleteTokenSub = dialogRef.afterClosed().subscribe((data) => {
            if (data?.confirmed) {
                this.createToken(data.expires, data.name, data.description)
            }
        })
    }

    createToken(expires: string, name: string, description : string): void {
        this._apiService
            .createAnalyticsToken(expires, name, description)
            .subscribe((newToken) => {
                this.analyticTokens = [...this.analyticTokens, newToken]
                this.getAnalyticsTokens()
                this._toastrService.success($localize`:@@tokenCreated: Token created`)
            })
    }

    copyTokenToClipboard(token: string): void {
        this._clipboard.copy(token)
        this._toastrService.info($localize`:@@copiedToClipboard: Copied to Clipboard`)
    }

    onDeleteTokenClick(tokenId: number): void {
        const dialogRef = this._dialog.open(DeleteDialog, {
            data: {
                headerText: $localize`:@@deleteSelectedToken:Delete selected token`
            }
        })

        this.createTokenSub = dialogRef.afterClosed().subscribe((confirmed) => {
            if (confirmed)
                this._apiService.deleteAnalyticsToken(tokenId).subscribe(() => {
                    this.getAnalyticsTokens()
                    this._toastrService.success($localize`:@@tokenDeleted: Token deleted`)
                })
        })
    }

    isVerbSelected(verb : AnalyticsTokenVerb, providerId : number) : boolean
    {
       return !!this.expandedElement?.analyticTokenVerbs?.find(v => v.verb == verb.id && providerId == v.provider) 
    }

    onSelectionChanged(verb : AnalyticsTokenVerb, providerId : number)
    {
        if(!this.expandedElement)
            return
        if(this.isVerbSelected(verb,providerId))
            this.expandedElement.analyticTokenVerbs = this.expandedElement.analyticTokenVerbs.filter(v => v.verb !== verb.id)
        else
            this.expandedElement.analyticTokenVerbs = [...this.expandedElement.analyticTokenVerbs, {...verb, verb : verb.id, id : "-1"}]
    }   

    saveActiveVerbs(): void {
        if(!this.expandedElement)
        return
        this._apiService.saveAnalyticTokenActiveVerbs(this.expandedElement.id, this.expandedElement.analyticTokenVerbs).subscribe(() => {
            this._toastrService.success($localize`:@@toastMessageSaved:Saved`)
            this.expandedElement = null
        })
    }

    isTokenExpired(expires: string): boolean {
        if (!expires) return false
        if (moment(expires) <= moment()) return true
        return false
    }

    fillCanAccessOptions(token: AnalyticsToken, options: AnalyticsToken[]): AnalyticsToken {
        return {
            ...token,
            can_access: options
                .filter((c) => c.id !== token.id)
                .map((b) => ({
                    ...b,
                    selected: !!token.can_access.find((a) => a.id === b.id)
                }))
        }
    }

    saveEngineResultAccess(id: AnalyticsTokenId): void {
    
        this._apiService
            .saveAccessRelation(
                id,
                this.expandedElement?.can_access.filter((e) => !!e.selected).map((e) => e.id) ?? []
            )
            .subscribe((data: any) => {
                this._toastrService.success($localize`:@@toastMessageSaved:Saved`)

                this.analyticTokens.map((e) => {
                    if (e.id === id) return this.fillCanAccessOptions(data, this.analyticTokens)
                    else return e
                })
            })
            
    }

    onUploadButtonClick()
    {
        if(!this.selectedFile)
            this.fileInputRef.nativeElement.click()
        else
            this.uploadSelectedFile()   
    }

    uploadFileChanged(event :Event)
    {
       this.selectedFile = (event.target as HTMLInputElement).files![0];
    }

    uploadSelectedFile()
    {
        this._apiService.uploadImageForAnalyticToken(
            this.expandedElement!.id,
            this.selectedFile!
        )
        .subscribe((data: any) => {
            this._toastrService.success($localize`:@@toastMessageSaved:Saved`)
            this.getAnalyticsTokens()
        })
    }

    getImageSource()
    {
        return `${environment.apiUrl}`+"/api/v1/provider/analytics-tokens/"+this.expandedElement?.id+"/image/"+this.expandedElement?.image_path
    }
}
