import hashlib
import json
import os
from venv import logger

import requests
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.http.response import JsonResponse
from django.shortcuts import render
from jsonschema import ValidationError, validate
from rest_framework import status
from rest_framework.views import APIView

from backend.utils import lrs_db
from consents.models import UserConsents
from providers.models import Provider, ProviderAuthorization, ProviderSchema
from users.models import CustomUser

PROJECT_PATH = os.path.abspath(os.path.dirname(__name__))

with open(os.path.join(PROJECT_PATH, "static/xapi_statement.schema.json")) as f:
    schema = json.load(f)


def store_in_db(x_api_statement):
    collection = lrs_db["statements"]
    try:
        result = collection.insert_one(x_api_statement)
        return str(result.inserted_id)
    except Exception as e:
        print("Failed to store xAPI statement in DB. ", e)
        return None


def anonymize_statement(x_api_statement):
    mbox = dict(
        enumerate(x_api_statement.get("actor", {}).get("mbox", "").split("mailto:"))
    ).get(1)
    account_email = (
        x_api_statement.get("actor", {}).get("account", {}).get("name", None)
    )
    email = mbox if mbox else account_email

    hashed_actor = hashlib.sha3_512(email.encode('utf-8')).hexdigest()

    x_api_statement.set("actor", {"name": "anonymous", "mbox":  settings.ANON_HASH_PREFIX + ": " + hashed_actor})


def process_statement(x_api_statement, provider, latest_schema):
    """
    Process xAPI statement by checking for validation errors and user consent settings.
    """
    try:
        validate(x_api_statement, schema)
    except ValidationError as e:
        return {"valid": False, "accepted": False, "reason": e.message}

    # system statements are directly stored without checking for consent
    if x_api_statement.get("actor", {}).get("mbox", "").startswith("system:"):
        if int((x_api_statement.get("actor", {}).get("mbox", "").split("system:"))[1]) == provider.id:
            return {"valid": True, "accepted": True}
        else:
            return {"valid": False, "accepted": False, "reason": "Wrong provider ID"}

    mbox = dict(
        enumerate(x_api_statement.get("actor", {}).get("mbox", "").split("mailto:"))
    ).get(1)
    account_email = (
        x_api_statement.get("actor", {}).get("account", {}).get("name", None)
    )
    if mbox is None and account_email is None:
        return {
            "valid": False,
            "accepted": False,
            "reason": "User missing in statement",
        }
    email = mbox if mbox else account_email

    if (not "verb" in x_api_statement) or (not "id" in x_api_statement["verb"]):
        return {"valid": False, "accepted": False, "reason": "No verb given"}
    verb = x_api_statement["verb"]["id"]

    if (not "object" in x_api_statement) or (not "id" in x_api_statement["object"]):
        return {"valid": False, "accepted": False, "reason": "No object given"}

    # find user by email
    try:
        user = CustomUser.objects.get(email=email)
    except ObjectDoesNotExist:
        return {"valid": False, "accepted": False, "message": "User not found"}

    # essential verbs do not require consent
    if not verb in [verb["id"] for verb in latest_schema.essential_verbs]:

        anon_verbs = [verb["id"] for verblist in [group["verbs"] for group in latest_schema.groups] for verb in verblist if verb.get("allowAnonymizedCollection", False)]

        # has the user paused data collection altogether?
        if user.paused_data_recording:
            return {
                "valid": True,
                "accepted": False,
                "reason": "User has paused data collection",
            }

        # has the user given consent to this verb?
        # maybe TODO: load correct provider schema pertaining to this user consent to validate the verb and objects fully
        user_consent = UserConsents.objects.filter(
            user=user, provider=provider, verb=verb, consented=True
        ).first()

        if not user_consent:
            if verb in anon_verbs:
                return {
                    "valid": True,
                    "accepted": True,
                    "needs_anonymization": True,
                }

            return {
                "valid": True,
                "accepted": False,
                "reason": f"User has not given consent to verb id {verb}",
            }

        # has the user given consent to the object at hand?
        object_consent = False
        verb_objects = json.loads(user_consent.object)
        object_statement = x_api_statement["object"]["definition"]
        if not verb_objects or len(verb_objects) == 0:  # no objects defined, "consent" by default
            object_consent = True
        else:
            for obj in verb_objects:
                if (
                    obj["consented"]
                    and obj["matching"] == "id"
                    and obj["id"] == x_api_statement["object"]["id"]
                ):
                    object_consent = True
                elif (
                    obj["consented"]
                    and obj["matching"] == "definitionType"
                    and obj["definition"]["type"]
                    == x_api_statement["object"]["definition"]["type"]
                ):
                    object_consent = True

        if not object_consent:
            if verb in anon_verbs:
                return {
                    "valid": True,
                    "accepted": True,
                    "needs_anonymization": True,
                }
            return {
                "valid": True,
                "accepted": False,
                "reason": f"User has not given consent to this object: {object_statement} for verb {verb}",
            }

        return {"valid": True, "accepted": True}

def process_tan_statement(x_api_statement):
    """
    Process xAPI statement by checking for validation errors.
    """
    try:
        validate(x_api_statement, schema)
    except ValidationError as e:
        return {"valid": False, "reason": e.message}


    # expects the tan to be in a field of the actor named "tan"
    tan = x_api_statement.get("actor", {}).get("tan", None)

    if tan is None:
        return {
            "valid": False,
            "reason": "TAN missing in statement",
        }

    if (not "verb" in x_api_statement) or (not "id" in x_api_statement["verb"]):
        return {"valid": False, "reason": "No verb given"}
    verb = x_api_statement["verb"]["id"]

    if (not "object" in x_api_statement) or (not "id" in x_api_statement["object"]):
        return {"valid": False, "reason": "No object given"}

    return {"valid": True}


class CreateStatement(APIView):
    """
    xAPI create statements proxy. This endpoint filter xAPI statements based and user settings and passes only consented xAPI statements to the LRS.
    """

    def post(self, request):
        # provider authorization
        auth_header = request.headers.get("Authorization")
        if not auth_header.startswith("Basic "):
            return JsonResponse(
                {
                    "message": "No provider authorization token supplied.",
                    "provider": "not found",
                },
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
        auth_key = auth_header.split(" ")[1]
        try:
            provider_auth = ProviderAuthorization.objects.get(key=auth_key)
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "No provider for authorization token found.",
                    "provider": "not found",
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )

        provider = provider_auth.provider

        # load latest provider schema for essential verbs
        try:
            latest_schema = ProviderSchema.objects.get(
                provider=provider, superseded_by__isnull=True
            )
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "No consent provider schema found.",
                    "provider": provider.name,
                },
                safe=False,
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

        # handle list of xAPI statements as well as single xAPI statement
        x_api_statements = (
            request.data if isinstance(request.data, list) else [request.data]
        )

        # TODO remove this actor override
        for stmt in x_api_statements:
            stmt["actor"]["account"]["name"] = "user1@polaris.com"

        if settings.SHOW_XAPI_STATEMENTS:
           print(x_api_statements)

        result = [
            process_statement(stmt, provider, latest_schema)
            for stmt in x_api_statements
        ]

        invalid_or_not_consented = (
            len([e for e in result if e["valid"] == False or e["accepted"] == False])
            > 0
        )        
        
        if settings.SHOW_XAPI_STATEMENTS:
            print(result)

        if invalid_or_not_consented:
            return JsonResponse(
                {
                    "message": "xAPI statements couldn't be stored in LRS",
                    "data": result,
                },
                status=status.HTTP_400_BAD_REQUEST,
            )
        else:
            # anonymize statements
            x_api_statements = [anonymize_statement(statement) if result[i].get("needs_anonymization", False)
                                else statement for i, statement in enumerate(x_api_statements)]
            uuids = list(map(store_in_db, x_api_statements))
            return JsonResponse(
                {
                    "message": "xAPI statements successfully stored in LRS",
                    "data": uuids,
                },
                safe=False,
                status=status.HTTP_200_OK,
            )


class CreateTANStatement(APIView):
    """
    xAPI create statements proxy for use with TANs. Stores all statements which contain TANs and are authorized with a provider token.
    """

    def post(self, request):
        # provider authorization
        auth_header = request.headers.get("Authorization")
        if not auth_header.startswith("Basic "):
            return JsonResponse(
                {
                    "message": "No provider authorization token supplied.",
                    "provider": "not found",
                },
                safe=False,
                status=status.HTTP_401_UNAUTHORIZED,
            )
        auth_key = auth_header.split(" ")[1]
        try:
            provider_auth = ProviderAuthorization.objects.get(key=auth_key)
        except ObjectDoesNotExist:
            return JsonResponse(
                {
                    "message": "No provider for authorization token found.",
                    "provider": "not found",
                },
                safe=False,
                status=status.HTTP_400_BAD_REQUEST,
            )

        provider = provider_auth.provider

        # handle list of xAPI statements as well as single xAPI statement
        x_api_statements = (
            request.data if isinstance(request.data, list) else [request.data]
        )

        result = [
            process_tan_statement(stmt)
            for stmt in x_api_statements
        ]

        invalid = (
            len([e for e in result if e["valid"] == False])
            > 0
        )

        if invalid:
            return JsonResponse(
                {
                    "message": "xAPI statements couldn't be stored in LRS",
                    "data": result,
                },
                status=status.HTTP_400_BAD_REQUEST,
            )
        else:
            # store provider as well, so statements can be properly assigned when searched by users
            for index, statement in enumerate(x_api_statements):
                statement["actor"]["provider_id"] = provider.id
                x_api_statements[index] = statement

            uuids = list(map(store_in_db, x_api_statements))
            return JsonResponse(
                {
                    "message": "xAPI statements successfully stored in LRS",
                    "data": uuids,
                },
                safe=False,
                status=status.HTTP_200_OK,
            )
